/* @(#)restore.c	1.24 04/10/05 Copyright 2003-2004 J. Schilling */
#ifndef lint
static	char sccsid[] =
	"@(#)restore.c	1.24 04/10/05 Copyright 2003-2004 J. Schilling";
#endif
/*
 *	Data base management for incremental restores
 *	needed to detect and execute rename() and unlink()
 *	operations between two incremental dumps.
 *
 *	Copyright (c) 2003-2004 J. Schilling
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 *	Problems:
 *
 *	-	lost+found will usually not extracted and thus
 *		get no new inode number.
 *		**** Solution: Use -U ****
 *
 *	-	chmod(1) only changes st_ctime but star will
 *		extract neither data nor permissions if only
 *		a chmod(1), chown(1) or smilar change did ocur.
 *		**** Solution: Use -U ****
 *
 *	-	unlink(dev) .... create(dev) and get same inode #
 *		**** Solution: sym_typecheck() ****
 *		The Solution creates a new problem that the new
 *		node created after sym_typecheck() may have a
 *		different inode number and we need to change the node.
 */

#include <mconfig.h>
#include <stdio.h>
#include <stdxlib.h>
#include <unixstd.h>
#include <standard.h>
#include "star.h"
#include "props.h"
#include "table.h"
#include "diff.h"
#include <strdefs.h>
#include <schily.h>
#include <dirdefs.h>	/* XXX Wegen S_IFLNK */
#include <statdefs.h>	/* XXX Wegen S_IRWXU */
#include "restore.h"
#include "starsubs.h"

extern	BOOL	debug;
extern	int	xdebug;
extern	int	verbose;
extern	dev_t	curfs;
extern	BOOL	forcerestore;

#define	HASH_ENTS	1024

LOCAL	imap_t	*iroot = NULL;		/* The root directory		  */
LOCAL	imap_t	*itmp = NULL;		/* The temp directory star-tmpdir */
LOCAL	imap_t	*isym = NULL;		/* The temp directory star-symtable */
LOCAL	imap_t	*imaps = NULL;
LOCAL	imap_t	**himaps = NULL;
LOCAL	imap_t	**hoimaps = NULL;
LOCAL	imap_t	**hnimaps = NULL;
#ifdef	DEBUG
LOCAL	int	nimp;
#endif

LOCAL	char	sym_tmpdir[] = "star-tmpdir";
LOCAL	char	sym_symtable[] = "star-symtable";

LOCAL	int	hashval		__PR((Uchar *str));
LOCAL	imap_t	*nfind_node	__PR((char *name, imap_t *idir));
LOCAL	imap_t	*pfind_node	__PR((char *name));
LOCAL	imap_t	*padd_node	__PR((char *name, ino_t oino, ino_t nino, Int32_t flags));
LOCAL	imap_t	*oifind_node	__PR((imap_t *idir, ino_t nino));
LOCAL	imap_t	*nifind_node	__PR((imap_t *idir, ino_t nino));
LOCAL	imap_t	*add_node	__PR((imap_t *cwd, char *name, ino_t oino, ino_t nino, Int32_t flags));
EXPORT	imap_t	*sym_addrec	__PR((FINFO *info));
EXPORT	void	sym_addstat	__PR((FINFO *info, imap_t *imp));
EXPORT	void	sym_dirprerpare	__PR((FINFO *info, imap_t *idir));
LOCAL	void	move2tmp	__PR((char *dir, char *name, ino_t oino, ino_t nino));
LOCAL	void	move2dir	__PR((char *dir, char *name, ino_t oino));
EXPORT	void	sym_typecheck	__PR((FINFO *info, FINFO *cinfo));
LOCAL	void	removefile	__PR((char *name));
EXPORT	void	sym_open	__PR((void));
LOCAL	void	sym_initsym	__PR((void));
LOCAL	void	tmpnotempty	__PR((void));
LOCAL	void	purgeent	__PR((imap_t *imp));
LOCAL	void	purgetree	__PR((imap_t *imp));
EXPORT	void	sym_init	__PR((GINFO *gp));
LOCAL	int	ngetline	__PR((FILE *f, char *buf, int len));
EXPORT	void	sym_close	__PR((void));
LOCAL	char	*fullname	__PR((imap_t *imp, char *cp, char *ep, BOOL top));
LOCAL	void	printfullname	__PR((FILE *f, imap_t *imp));
LOCAL	void	printonesym	__PR((FILE *f, imap_t *imp));
LOCAL	void	printsyms	__PR((FILE *f, imap_t *imp));
#ifdef	__needed__
LOCAL	BOOL	dirdiskonly	__PR((FINFO *info, int *odep, char ***odp));
#endif

LOCAL int
hashval(str)
	register Uchar	*str;
{
	register int	sum = 0;
	register int	i;
	register int	c;

	for (i = 0; (c = *str++) != '\0'; i++)
		sum ^= (c << (i&7));
	return (sum);
}

/*
 * Find a node by path name component and directory node pointer
 */
LOCAL imap_t *
nfind_node(name, idir)
	char	*name;
	register imap_t	*idir;
{
	register imap_t	*imp;
	register char	*np;
	register char	*inp;
	register int	nhash;

	np = name;
	if (np[0] == '\0')
		return (idir);
	if (np[0] == '.') {
		if (np[1] == '\0')
			return (idir);
		if (np[1] == '.' && np[2] == '\0')
			return (idir->i_dparent);
	}
	nhash = hashval((Uchar *)name);
	imp = himaps[nhash%HASH_ENTS];

	for (; imp; imp = imp->i_hnext) {

		if (nhash != imp->i_hash)
			continue;
		if (imp->i_dparent != idir)
			continue;
		if (imp->i_flags & I_DELETE)
			continue;

		np = name;
		inp = imp->i_name;
		while (*inp) {
			if (*np++ != *inp++)
				goto nextmatch;
		}
		if (np[0] == '\0')
			break;
nextmatch:
		;
	}
	return (imp);
}

/*
 * Find a node by full path name
 */
LOCAL imap_t *
pfind_node(name)
	char	*name;
{
	register imap_t	*imp = iroot;
	register char	*np = name;
	register char	*sp;

	if (imp == NULL)
		return (imp);

	while (*np) {
		if ((sp = strchr(np, '/')) != NULL)
			*sp = '\0';
		if ((imp->i_flags & I_DIR) == 0) {
			errmsgno(EX_BAD, "pfind_node(): Not a directory '%s' for '%s'.\n", imp->i_name, name);
			return (NULL);
		}
		imp = nfind_node(np, imp);
		if (sp)
			*sp = '/';
		else
			break;
		np = &sp[1];
		if (imp == NULL)
			break;
	}
	return (imp);
}

/*
 * Add a new full path name and return a node pointer
 */
LOCAL imap_t *
padd_node(name, oino, nino, flags)
	char	*name;
	ino_t	oino;
	ino_t	nino;
	Int32_t	flags;
{
	register imap_t	*imp = iroot;
	register imap_t	*idir;
	register char	*np = name;
	register char	*sp;

	if (imp == NULL) {
		if ((np[0] == '.' && np[1] == '\0') ||
		    (np[0] == '.' && np[1] == '/' && np[2] == '\0')) {
			imp = add_node(NULL, ".", oino, nino, flags);
			imp->i_dparent = imp;
			iroot = imp;
			return (imp);
		}
		comerrno(EX_BAD, "Panic no root to add '%s'.\n", name);
	}

	while (*np) {
		if ((sp = strchr(np, '/')) != NULL)
			*sp = '\0';

		if ((imp->i_flags & I_DIR) == 0) {
			errmsgno(EX_BAD, "padd_node(): Not a directory '%s' for '%s'.\n", imp->i_name, name);
			return (NULL);
		}
		idir = imp;
		imp = nfind_node(np, idir);
		if (imp == NULL) {
			if (sp != NULL && strchr(&sp[1], '/') != NULL)
				imp = add_node(idir, np, 0, 0, I_DIR);
			else
				imp = add_node(idir, np, oino, nino, flags);
		}
		if (sp)
			*sp = '/';
		else
			break;
		np = &sp[1];
	}
	if (imp->i_oino == 0 && oino != 0) {
		int	hv;

		imp->i_oino = oino;
		hv = oino % HASH_ENTS;
		imp->i_honext = hoimaps[hv];
		hoimaps[hv] = imp;
	}
	if (imp->i_nino == 0 && nino != 0) {
		int	hv;

		imp->i_nino = nino;
		hv = nino % HASH_ENTS;
		imp->i_hnnext = hnimaps[hv];
		hnimaps[hv] = imp;
	}
	return (imp);
}

/*
 * Find a node by old inode number
 */
LOCAL imap_t *
oifind_node(idir, oino)
	imap_t	*idir;
	ino_t	oino;
{
	register imap_t	*imp;

	imp = hoimaps[oino%HASH_ENTS];

	for (; imp; imp = imp->i_honext) {

		if (imp->i_flags & I_DELETE)
			continue;
		if (idir && idir != imp->i_dparent)
			continue;

		/*
		 * XXX Dies ist eine halbfertige Directory.
		 * XXX Hoffentlich brauchen wir diesen Knoten nie.
		 */
		if (imp->i_nino == 0)
			continue;
		if (oino == imp->i_oino)
			break;
	}
	return (imp);
}

/*
 * Find a node by new inode number
 */
LOCAL imap_t *
nifind_node(idir, nino)
	imap_t	*idir;
	ino_t	nino;
{
	register imap_t	*imp;

	imp = hnimaps[nino%HASH_ENTS];

	for (; imp; imp = imp->i_hnnext) {

		if (imp->i_flags & I_DELETE)
			continue;
		if (idir && idir != imp->i_dparent)
			continue;
		if (nino == imp->i_nino)
			break;
	}
	return (imp);
}

/*
 * Add a new node by path name component and directory node pointer
 * return a node pointer
 */
LOCAL imap_t *
add_node(cwd, name, oino, nino, flags)
	imap_t	*cwd;
	char	*name;
	ino_t	oino;
	ino_t	nino;
	Int32_t	flags;
{
	register imap_t	*imp;
	register int	hv;

	imp = __malloc(sizeof (imap_t), "new imap");
	imp->i_name = __savestr(name);
	imp->i_hash = hashval((Uchar *)imp->i_name);
	imp->i_oino = oino;
	imp->i_nino = nino;
	imp->i_flags = flags;
	imp->i_next = imaps;
	imaps = imp;

	imp->i_dir = NULL;
	imp->i_dnext = NULL;
	imp->i_dxnext = NULL;
	imp->i_dparent = cwd;
	if (cwd) {
		imp->i_dxnext = cwd->i_dir;
		cwd->i_dir = imp;
	}

	hv = imp->i_hash % HASH_ENTS;
	imp->i_hnext = himaps[hv];
	himaps[hv] = imp;

	if (oino) {
		hv = oino % HASH_ENTS;
		imp->i_honext = hoimaps[hv];
		hoimaps[hv] = imp;
	} else {
		imp->i_honext = NULL;
	}
	if (nino) {
		hv = nino % HASH_ENTS;
		imp->i_hnnext = hnimaps[hv];
		hnimaps[hv] = imp;
	} else {
		imp->i_hnnext = NULL;
	}
#ifdef	DEBUG
	nimp++;
#endif

	return (imp);
}

/*
 * Add a record for a file. If this is a directory, add entries for the
 * directory content.
 */
EXPORT imap_t *
sym_addrec(info)
	FINFO	*info;
{
	imap_t	*imp;
	BOOL	isold = TRUE;

	if ((imp = pfind_node(info->f_name)) == NULL) {
		isold = FALSE;
		imp = padd_node(info->f_name, info->f_ino, (ino_t)0, 0);
	}

	if (is_dir(info)) {
		imp->i_flags |= I_DIR;
	}

	if (is_dir(info)) {
		imap_t	*idir;
		char	*dp;
		ino_t	*ip;
		int	len;
		int	i;

		idir = imp;
		dp = info->f_dir;
		ip = info->f_dirinos;
		if (dp == NULL || ip == NULL)
			return (idir);

		i = 0;
		while (*dp) {
			dp++;
			len = strlen(dp);

			if ((imp = nfind_node(dp, idir)) == NULL) {
				imp = add_node(idir, dp, (ino_t)0, (ino_t)0, 0);
			}
			if (imp->i_oino == (ino_t)0) {
				int	hv;

				imp->i_oino = ip[i];
				hv = imp->i_oino % HASH_ENTS;
				imp->i_honext = hoimaps[hv];
				hoimaps[hv] = imp;
			}
			i++;
			dp += len+1;
		}
		return (idir);
	}
	return (imp);
}

/*
 * Add status information for a node.
 */
EXPORT void
sym_addstat(info, imp)
	FINFO	*info;
	imap_t	*imp;
{
	FINFO	finfo;
	BOOL	isold = TRUE;
	int	hv;

	if (imp == NULL && (imp = pfind_node(info->f_name)) == NULL) {
		isold = FALSE;
		imp = padd_node(info->f_name, info->f_ino, (ino_t)0, 0);
	}
	if (imp->i_oino == 0) {
		if (imp != iroot) {
			errmsgno(EX_BAD,
			"WARNING: late old inode add for '%s'\n", info->f_name);
		}
		imp->i_oino = info->f_ino;
		hv = imp->i_oino % HASH_ENTS;
		imp->i_honext = hoimaps[hv];
		hoimaps[hv] = imp;
	}

	fillbytes((char *)&finfo, sizeof (finfo), '\0');
	_getinfo(info->f_name, &finfo);
	if (imp->i_nino != 0 && imp->i_nino != finfo.f_ino)
		errmsgno(EX_BAD, "%s nino change from %lld to %lld\n",
			info->f_name,
			(Llong)imp->i_nino, (Llong)finfo.f_ino);

	/*
	 * An nino change implies a hash change. We need to remove the node
	 * from the old hash list and put it into the new list.
	 */
	if (imp->i_nino != finfo.f_ino) {
		BOOL	newhash = FALSE;

		if ((imp->i_nino != 0) &&
		    ((imp->i_nino % HASH_ENTS) != (finfo.f_ino % HASH_ENTS))) {
			imap_t	*tnp;

			tnp = hnimaps[imp->i_nino % HASH_ENTS];
			for (; tnp; tnp = tnp->i_hnnext) {
				if (imp == tnp->i_hnnext) {
					tnp->i_hnnext = imp->i_hnnext;
					break;
				}
			}
			newhash = TRUE;
		}

		if (imp->i_nino == 0 || newhash) {
			hv = finfo.f_ino % HASH_ENTS;
			imp->i_hnnext = hnimaps[hv];
			hnimaps[hv] = imp;
		}
		imp->i_nino = finfo.f_ino;
	}

	if (is_dir(info)) {
		int	len = strlen(imp->i_name);

		if (imp->i_name[len-1] == '/')
			imp->i_name[len-1] = '\0';
		imp->i_flags |= I_DIR;
	}
}

/*
 * Prepare the directory for the following extraction.
 * This is done by first removing all files that are no longer present on the
 * current incremental and then trying to move/link missing files that are
 * a result of a rename or link operation.
 */
EXPORT void
sym_dirprepare(info, idir)
	FINFO	*info;
	imap_t	*idir;
{
	int	dlen;
	ino_t	*oino;
	char	**dname;
	char	*p;
	int	i;
	int	j;
	FINFO	finfo;
	char	*dp2;
	int	ents2;
	char	**dname2;
	ino_t	*ino2;
	ino_t	*oino2;

	if (info->f_dir == NULL) {
		/*
		 * Most likely a mount point
		 */
		if (info->f_dev == curfs)
			comerrno(EX_BAD, "Archive contains directory '%s' without name list\n", info->f_name);
		return;
	}

	/*
	 * Check if the node on disk is really a directory.
	 * As we did already check the directory content of the higher direcory
	 * it should never happen that the on disk node is not a directory.
	 */
	fillbytes((char *)&finfo, sizeof (finfo), '\0');
	if (!_getinfo(info->f_name, &finfo))
		return;
	if (!is_dir(&finfo))
		comerrno(EX_BAD, "Panic: '%s' is not a directory.\n", info->f_name);

	if ((dname = __malloc(info->f_dirents * sizeof (char *), "sym_dirprepare name")) == NULL) {
		return;
	}
	/*
	 * Put directory names from archive into array.
	 */
	p = info->f_dir;
	oino = info->f_dirinos;
	dlen = info->f_dirents;
	for (i = 0; i < dlen; i++) {
		dname[i] = &p[1];
		p += strlen(&p[1]) + 2;
	}

#ifdef	RES_DEBUG
	for (i = 0; i < dlen; i++) {
		error("INO %lld NAME %s\n",
			(Llong)info->f_dirinos[i],
			dname[i]);
	}
#endif

	dp2 = fetchdir(info->f_name, &ents2, 0, &ino2);
	if ((oino2 = __malloc(ents2 * sizeof (ino_t), "sym_dirprepare oino2")) == NULL) {
		return;
	}
	if ((dname2 = __malloc(ents2 * sizeof (char *), "sym_dirprepare name2")) == NULL) {
		return;
	}
	/*
	 * Put directory names from disk into array.
	 */
	p = dp2;
	for (i = 0; i < ents2; i++) {
		dname2[i] = &p[1];
		p += strlen(&p[1]) + 2;
	}
	for (i = 0; i < ents2; i++) {
		imap_t	*imp;
		imp = nifind_node((imap_t *)0, ino2[i]);
		if (imp == NULL) {
			errmsgno(EX_BAD,
			"Panic: No symbol entry for inode %lld (%s).\n",
			(Llong)ino2[i], dname2[i]);

			oino2[i] = 0;
			continue;
		}
		if (imp->i_oino == (ino_t)0) {
			if (imp != itmp && imp != isym) {
				errmsgno(EX_BAD,
				"Panic: No old inode for inode %lld (%s).\n",
				(Llong)ino2[i], dname2[i]);
			}
		}
		oino2[i] = imp->i_oino;
	}

	if (xdebug)
		error("sym_dirprepare(%s)\n", info->f_name);
	/*
	 * Check for all files that are in the current on disk directory
	 * but no longer exist on the current incremental with the same
	 * name and inode number as before.
	 */
	for (i = 0; i < ents2; i++) {
		ino_t	in;
		if (oino2[i] == 0)
			continue;
		in = oino2[i];

		if (xdebug)
			error("Checking ino %lld (%s)...", (Llong)oino2[i], dname2[i]);
		for (j = 0; j < dlen; j++) {
/*error("in %lld oino %lld\n", (Llong)in, (Llong)oino[j]);*/
			if (in == oino[j] && streql(dname2[i], dname[j]))
				break;
		}
		if (j >= dlen) {
			if (xdebug)
				error("RM\n");
				/* dir name, file name,  old inode#, new inode # */
			move2tmp(info->f_name, dname2[i], in, ino2[i]);
		} else {
			if (xdebug)
				error("found\n");
		}
	}

	if (xdebug)
		error("sym_dirprepare(%s) <<<<\n", info->f_name);
	/*
	 * Check for all files that are on the current incremental but
	 * are not in the current on disk directory  with the same
	 * and inode number.
	 */
	for (i = 0; i < dlen; i++) {
		ino_t	in;
		in = oino[i];

		if (xdebug)
			error("Checking ino %lld (%s)...", (Llong)oino[i], dname[i]);
		for (j = 0; j < ents2; j++) {
			if (in == oino2[j] && streql(dname2[j], dname[i]))
				break;
		}
		if (j >= ents2) {
			if (xdebug)
				error("LINK\n");
				/* dir name, file name,  old inode# */
			move2dir(info->f_name, dname[i], in);
		} else {
			if (xdebug)
				error("found\n");
		}
	}

	if (dname != NULL)
		free(dname);
	if (oino2 != NULL)
		free(oino2);
	if (dname2 != NULL)
		free(dname2);
	if (dp2 != NULL)
		free(dp2);
	if (ino2 != NULL)
		free(ino2);
}

/*
 * Move a file to the temp directory. This file has been identified
 * to be no longer in the current directory pointed to by 'dir'.
 */
LOCAL void
move2tmp(dir, name, oino, nino)
	char	*dir;
	char	*name;
	ino_t	oino;
	ino_t	nino;
{
	char	path[2*PATH_MAX+1];
	char	tpath[128];
	imap_t	*onp;
	imap_t	*nnp;

	js_snprintf(path, sizeof (path), "%s/%s", dir, name);
	js_snprintf(tpath, sizeof (tpath), "%s/#%lld", sym_tmpdir, (Llong)nino);

	onp = pfind_node(path);
	if (onp == NULL) {
		comerrno(EX_BAD,
			"Panic: amnesia in inode data base for '%s'.\n",
			path);
	}
	nnp = nifind_node(itmp, nino);
	if (nnp) {			/* inode is already in star-tmpdir */
		if (xdebug)
			error("unlink(%s)\n", path);
		if (unlink(path) < 0)
			comerr("Cannot unlink '%s'.\n", path);
		purgeent(onp);
		return;
	}

	if (xdebug)
		error("rename(%s, %s)\n", path, tpath);
	if (rename(path, tpath) < 0)
		comerr("Cannot rename '%s' to '%s'.\n", path, tpath);

	nnp = padd_node(tpath, onp->i_oino, onp->i_nino, onp->i_flags);
	if (xdebug)
		error("move2tmp() itmp %p onp %p nnp %p\n", itmp, onp, nnp);

	purgeent(onp);
	nnp->i_dir = onp->i_dir;
	for (onp = nnp->i_dir; onp; onp = onp->i_dxnext) {
		onp->i_dparent = nnp;
	}
}

/*
 * Try to move a file from the temp directory to the current directory
 * pointed to by 'dir'.
 */
LOCAL void
move2dir(dir, name, oino)
	char	*dir;
	char	*name;
	ino_t	oino;
{
	char	path[2*PATH_MAX+1];
	char	tpath[2*PATH_MAX+1];
	imap_t	*onp;
	imap_t	*nnp;
	char	*p;

	js_snprintf(path, sizeof (path), "%s/%s", dir, name);
	onp = oifind_node(itmp, oino);

#ifdef	RES_DEBUG
	if (onp)
		error("move2dir(%s, %s, %lld) = %s\n",
			dir, name, (Llong)oino, onp->i_name);
	else
		error("move2dir(%s, %s, %lld) = NOT FOUND\n",
			dir, name, (Llong)oino);
#endif

	if (onp) {
		js_snprintf(tpath, sizeof (tpath), "%s/%s",
						sym_tmpdir, onp->i_name);
		if (xdebug)
			error("rename(%s, %s)\n", tpath, path);
		if (rename(tpath, path) < 0) {
			errmsg("Cannot rename '%s' to '%s'.\n", tpath, path);
			/* XXX set error code */
			return;
		}
		nnp = padd_node(path, onp->i_oino, onp->i_nino, onp->i_flags);
		if (xdebug) {
			error("move2dir() itmp %p onp %p nnp %p\n",
							itmp, onp, nnp);
		}
		purgeent(onp);
		nnp->i_dir = onp->i_dir;
		for (onp = nnp->i_dir; onp; onp = onp->i_dxnext) {
			onp->i_dparent = nnp;
		}
		return;
	}

	/*
	 * XXX Wenn nicht in star-tmpdir, dann
	 *
	 * -	hard link zu File versuchen
	 *
	 * -	rename einer Directory versuchen und Flag setzen
	 *
	 * -	Ist Flag gesetzt, dann Hard Link auf Dir setzen.
	 */
	if (xdebug)
		error("Cannot rename '%s' from '%s'.\n", path, sym_tmpdir);

	onp = oifind_node((imap_t *)0, oino);
	if (onp == NULL) {
		if (xdebug)
			error("Cannot link/move any file to '%s'.\n",
								path);
		return;
	}
	p = fullname(onp, tpath, &tpath[sizeof (tpath)], TRUE);
	if (p == NULL) {
		/* XXX error code */
		errmsgno(EX_BAD, "Path name '");
		printfullname(stderr, onp);
		error("' too long, cannot rename\n");
		return;
	}
	if (xdebug)
		error("move2dir(%s, %s, %lld) found path => '%s'\n",
			dir, name, (Llong)oino, tpath);

	if ((onp->i_flags & (I_DIR|I_DID_RENAME)) == I_DIR) {
		if (xdebug)
			error("rename(%s, %s)\n", tpath, path);
		if (rename(tpath, path) < 0) {
			/* XXX error code */
			errmsg("Cannot rename(%s, %s)\n", tpath, path);
		} else {
			nnp = padd_node(path, onp->i_oino, onp->i_nino, onp->i_flags);
			purgeent(onp);
			nnp->i_flags |= I_DID_RENAME;
			nnp->i_dir = onp->i_dir;
			for (onp = nnp->i_dir; onp; onp = onp->i_dxnext) {
				onp->i_dparent = nnp;
			}
		}
		return;
	}
	if (xdebug)
		error("link(%s, %s)\n", tpath, path);
	if (link(tpath, path) < 0) {
		/* XXX error code */
		errmsg("Cannot link(%s, %s)\n", tpath, path);
	} else {
		nnp = padd_node(path, onp->i_oino, onp->i_nino, onp->i_flags);
#ifdef	nonono
		/*
		 * XXX It is not clear how to handle new hard links
		 * XXX to directories correctly.
		 */
		nnp->i_dir = onp->i_dir;
		for (onp = nnp->i_dir; onp; onp = onp->i_dxnext) {
			onp->i_dparent = nnp;
		}
#endif
	}
}

/*
 * Check whether the target file exists and has a
 * different type or is a dev node with different
 * major/minor numbers. In this case, we need to
 * remove the file. This happend when the original
 * file has been removed and a new (different) file
 * with the same name did get the same inode number.
 */
EXPORT void
sym_typecheck(info, cinfo)
	FINFO	*info;
	FINFO	*cinfo;
{
	fillbytes((char *)cinfo, sizeof (*cinfo), '\0');
	if (!_getinfo(info->f_name, cinfo))
		return;

	if (info->f_filetype == cinfo->f_filetype &&
	    !is_special(cinfo) && !is_symlink(cinfo))
		return;

	if (is_symlink(cinfo) || is_special(cinfo)) {
		removefile(info->f_name);
		return;
	}

	if (info->f_type == cinfo->f_type)
		return;

	if (is_dir(cinfo)) {
		imap_t	*onp;

		if ((onp = pfind_node(info->f_name)) != NULL) {
			imap_t	*imp;
			for (imp = onp->i_dir; imp; imp = imp->i_dxnext) {
				move2tmp(info->f_name, imp->i_name, imp->i_oino, imp->i_nino);
			}
		}
	}

	removefile(info->f_name);
}

LOCAL void
removefile(name)
	char	*name;
{
	extern	BOOL	force_remove;
		BOOL	of = force_remove;

	if (xdebug)
		error("REMOVE %s\n", name);

	force_remove = TRUE;
	(void) remove_file(name, TRUE);
	force_remove = of;
}

/*
 * Read in the old inode symbol table
 */
EXPORT void
sym_open()
{
	char	buf[2*PATH_MAX+1];
	FILE	*f = fileopen(sym_symtable, "r");
	int	amt;
	char	*p;
	Llong	ll;
	ino_t	oino;
	ino_t	nino;
	mode_t	old_umask;
	char	td[] = "star-tmpdir/.";
	imap_t	*ilast = NULL;
	imap_t	*icwd = NULL;
	FINFO	finfo;

#define	PERM_BITS	(S_IRWXU|S_IRWXG|S_IRWXO)	/* u/g/o basic perm */

	fillbytes((char *)&finfo, sizeof (finfo), '\0');
	if (_getinfo(sym_symtable, &finfo) && !is_file(&finfo)) {
		errmsgno(EX_BAD, "'%s' is not a file.\n", sym_symtable);
		comerrno(EX_BAD, "Remove '%s' and try again.\n", sym_symtable);
	}

	if (himaps == NULL) {
		register int	hv;

		himaps = __malloc(HASH_ENTS * sizeof (imap_t *), "imap hash");
		hoimaps = __malloc(HASH_ENTS * sizeof (imap_t *), "oimap hash");
		hnimaps = __malloc(HASH_ENTS * sizeof (imap_t *), "nimap hash");
		for (hv = 0; hv < HASH_ENTS; hv++) {
			himaps[hv] = 0;
			hoimaps[hv] = 0;
			hnimaps[hv] = 0;
		}
	}
	old_umask = umask((mode_t)0);
	umask(PERM_BITS & ~S_IRWXU);
	if (!create_dirs(td)) {
		/*
		 * This also fails if star-tmpdir is not a directory.
		 */
		comerrno(EX_BAD, "Cannot create '%s'.\n", sym_tmpdir);
	}
	umask(old_umask);

	if (f == NULL)
		return;

	/*
	 * XXX Hier mu eine berprfung rein fr:
	 * XXX -	star version
	 * XXX -	Dump level	(Archive vs. Disk)
	 * XXX -	Dump date	(Archive vs. Disk)
	 */
	while ((amt = ngetline(f, buf, sizeof (buf))) > 0) {
		Int32_t	flags;

		if (getc(f) != '\n')
			comerrno(EX_BAD, "Missing newline at end of record.\n");

		if (buf[0] == '>' && buf[1] == '\0') {
			if (iroot == NULL) {
				iroot = icwd = ilast;
				icwd->i_dparent = icwd;
				if (icwd->i_name[0] != '.' || icwd->i_name[1] != '\0')
					comerrno(EX_BAD, "Bad root dir '%s' in '%s'.\n",
						icwd->i_name,
						sym_symtable);
			} else {
				icwd->i_dnext = ilast;
				if (ilast->i_dparent != icwd)
					comerrno(EX_BAD, "Bad parent dir.\n");
				ilast->i_dparent = icwd;
				icwd = ilast;
			}
			continue;
		}
		if (buf[0] == '<' && buf[1] == '\0') {
			icwd = icwd->i_dparent;
			icwd->i_dnext = NULL;
			continue;
		}

		p = buf;
		flags = 0;
		if (*p == 'D')
			flags |= I_DIR;

		p++;	/* Skip Flag */
		p = astollb(p, &ll, 10);
		if (*p != '\t')
			comerrno(EX_BAD, "Missing TAB after old ino.\n");
		oino = ll;
		p = astollb(p, &ll, 10);
		if (*p != '\t')
			comerrno(EX_BAD, "Missing TAB after NEW ino.\n");
		nino = ll;
		ilast = add_node(icwd, ++p, oino, nino, flags);
#ifdef	__needed__
		if (0) {
		imap_t	*ip = iroot;

			for (; ip; ip = ip->i_dnext) {
				printf("%s/", ip->i_name);
				if (ip == icwd)
					break;
			}
			printf("%s\n", p);
		}
#endif
	}
	fclose(f);
#ifdef	__needed__
if (0)
printsyms(stderr, iroot);
#endif
}

/*
 * Make sure "star-tmpdir" exists and is part of the inode symbol cache.
 * Make sure that "star-symtable" is part of the inode symbol cache if
 * it exists.
 */
LOCAL void
sym_initsym()
{
	char	tpath[PATH_MAX+1];
	FINFO	finfo;
	imap_t	*imp;
	char	*dp;
	int	ents;

	fillbytes((char *)&finfo, sizeof (finfo), '\0');
	if (!_getinfo(sym_tmpdir, &finfo))
		comerr("Cannot stat '%s'\n", sym_tmpdir);

	if (iroot == NULL)
		imp = padd_node(".", (ino_t)0, (ino_t)0, I_DIR);

	if ((imp = pfind_node(sym_tmpdir)) == NULL)
		imp = padd_node(sym_tmpdir, (ino_t)0, finfo.f_ino, I_DIR);

	if (imp->i_nino != finfo.f_ino)
		errmsgno(EX_BAD, "%s nino change from %lld to %lld\n",
			sym_tmpdir,
			(Llong)imp->i_nino, (Llong)finfo.f_ino);

	imp->i_nino = finfo.f_ino;
	itmp = imp;

	for (imp = imp->i_dir; imp; imp = imp->i_dxnext) {
		if (imp->i_flags & I_DELETE)
			continue;

		js_snprintf(tpath, sizeof (tpath), "%s/%s",
						sym_tmpdir, imp->i_name);
		if (!_getinfo(tpath, &finfo))
			purgetree(imp);
		else
			tmpnotempty();
	}
	ents = 0;
	dp = fetchdir(sym_tmpdir, &ents, 0, (ino_t **)0);
	if (dp)
		free(dp);
	if (ents > 0)
		tmpnotempty();

	fillbytes((char *)&finfo, sizeof (finfo), '\0');
	if (_getinfo(sym_symtable, &finfo)) {
		if ((imp = pfind_node(sym_symtable)) == NULL)
			imp = padd_node(sym_symtable, (ino_t)0, finfo.f_ino, 0);
		isym = imp;
	}
}

LOCAL void
tmpnotempty()
{
	errmsgno(EX_BAD, "The directory '%s' is not empty.\n", sym_tmpdir);
	comerrno(EX_BAD, "Remove all files in '%s' and try again.\n", sym_tmpdir);
}

LOCAL void
purgeent(imp)
	imap_t	*imp;
{
	imp->i_flags |= I_DELETE;
	free(imp->i_name);
	imp->i_name = NULL;
}

LOCAL void
purgetree(imp)
	imap_t	*imp;
{
	if (imp == NULL)
		return;

	for (; imp; imp = imp->i_dxnext) {

		if (imp->i_flags & I_DELETE)
			continue;

		purgetree(imp->i_dir);
		purgeent(imp);
	}
}

/*
 * Do some checks and initialisations before the symbols are going to be used
 */
EXPORT void
sym_init(gp)
	GINFO	*gp;
{
/*extern	char	*vers;*/
/*error("Star version '%s'\n", vers);*/
/*	error("imaps: %p level %d\n", imaps, gp->dumplevel);*/

	/*
	 * XXX imaps -> iroot ???
	 */
	if (gp->dumplevel == 0 && imaps != NULL) {
		errmsgno(EX_BAD, "Trying to extract a level 0 dump but '%s' exists.\n", sym_symtable);
		comerrno(EX_BAD, "Remove '%s' and try again.\n", sym_symtable);
	}
	if (gp->dumplevel > 0 && imaps == NULL) {
		errmsgno(EX_BAD, "Trying to extract a level %d dump but '%s' does not exist.\n",
				gp->dumplevel, sym_symtable);
		comerrno(EX_BAD, "Restore the level 0 dump first and try again.\n");
	}
	if (gp->dumptype != DT_FULL) {
		errmsgno(EX_BAD, "WARNING: This dump is a %s dump and not a full dump.\n",
			dt_name(gp->dumptype));
		errmsgno(EX_BAD, "WARNING: An incremental restore may not work correctly.\n");
		if (!forcerestore) {
			comerrno(EX_BAD, "Use -force-restore if you want to restore anyway.\n");
			/* NOTREACHED */
		}
	}
	/*
	 * XXX Hier mu eine berprfung rein fr:
	 * XXX -	star version
	 * XXX -	Dump level	(Archive vs. Disk)
	 * XXX -	Dump date	(Archive vs. Disk)
	 */

	sym_initsym();
}

/*
 * A special version of fgetline() that does not stop on '\n' but only on '\0'.
 */
LOCAL int
ngetline(f, buf, len)
	register	FILE	*f;
			char	*buf;
	register	int	len;
{
	register int	c	= '\0';
	register char	*bp	= buf;
	register int	nul	= '\0';

	for (;;) {
		if ((c = getc(f)) < 0)
			break;
		if (c == nul)
			break;
		if (--len > 0) {
			*bp++ = (char)c;
		} else {
			/*
			 * Read up to end of line
			 */
			while ((c = getc(f)) >= 0 && c != nul)
				/* LINTED */
				;
			break;
		}
	}
	*bp = '\0';
	/*
	 * If buffer is empty and we hit EOF, return EOF
	 */
	if (c < 0 && bp == buf)
		return (c);

	return (bp - buf);
}

/*
 * Write back the inode symbol table
 */
EXPORT void
sym_close()
{
	char	tpath[PATH_MAX+1];
	FINFO	finfo;
	FILE	*f;
	imap_t	*imp;
	BOOL	warned = FALSE;

	f = filemopen(sym_symtable, "wct", S_IRUSR|S_IWUSR);
	if (f == NULL) {
		errmsg("Cannot create '%s'.\n", sym_symtable);
		return;
	}
	if ((imp = pfind_node(sym_symtable)) == NULL) {
		fillbytes((char *)&finfo, sizeof (finfo), '\0');
		if (!_getinfo(sym_symtable, &finfo))
			comerr("Cannot stat '%s'\n", sym_symtable);

		imp = padd_node(sym_symtable, (ino_t)0, finfo.f_ino, 0);
	}
	if (isym == NULL)
		isym = imp;

	if (itmp != NULL)
	for (imp = itmp->i_dir; imp; imp = imp->i_dxnext) {
		if (imp->i_flags & I_DELETE)
			continue;

		js_snprintf(tpath, sizeof (tpath), "%s/%s",
						sym_tmpdir, imp->i_name);
		fillbytes((char *)&finfo, sizeof (finfo), '\0');
		if (_getinfo(tpath, &finfo)) {
			if (is_dir(&finfo)) {
				if (rmdir(tpath) >= 0)
					purgeent(imp);
			} else {
				if (finfo.f_nlink > 1) {
					if (unlink(tpath) >= 0)
						purgeent(imp);
				}
			}
			if ((imp->i_flags & I_DELETE) == 0 && !warned) {
				errmsgno(EX_BAD,
				"Don't forget to remove the files in '%s'.\n",
				sym_tmpdir);
				warned = TRUE;
			}
		}
	}

	printsyms(f, iroot);
	fclose(f);
}

/*
 * Create a full path name for imp in cp
 */
LOCAL char *
fullname(imp, cp, ep, top)
	imap_t 	*imp;
	char	*cp;
	char	*ep;
	BOOL	top;
{
	int	len;

	if (imp == iroot)
		return (cp);

	cp = fullname(imp->i_dparent, cp, ep, FALSE);
	if (cp == NULL)
		return (NULL);

	len = strlen(imp->i_name) + (top ? 0:1);
	if (cp + len >= ep)
		return (NULL);

	strcpy(cp, imp->i_name);
	if (!top) {
		cp += len-1;
		*cp++ = '/';
		*cp   = '\0';
	} else {
		cp += len;
	}
	return (cp);
}

/*
 * Print a full path name by parsing the tree of parent directories
 */
LOCAL void
printfullname(f, imp)
	FILE	*f;
	imap_t 	*imp;
{
	if (imp == iroot)
		return;
	printfullname(f, imp->i_dparent);
	fprintf(f, "/%s", imp->i_name);
}

/*
 * Print a single symbol to a file
 */
LOCAL void
printonesym(f, imp)
	FILE	*f;
	imap_t 	*imp;
{
	/*	Flags	Old ino	New Ino		Name */
	fprintf(f, "%c %lld	%lld	%s%c\n",
			imp->i_flags & I_DIR ? 'D':' ',
			(Llong)imp->i_oino, (Llong)imp->i_nino,
			imp->i_name,
			0);

	if (imp->i_oino == 0) {
		if (imp != itmp && imp != isym) {
			errmsgno(EX_BAD, "WARNING: No old inode number for ");
			printfullname(stderr, imp);
			fprintf(stderr, "\n");
		}
	}
	if (imp->i_nino == 0) {
		errmsgno(EX_BAD, "WARNING: No new inode number for ");
		printfullname(stderr, imp);
		fprintf(stderr, "\n");
	}
}

/*
 * Print all symbols to a file
 */
LOCAL void
printsyms(f, imp)
	FILE	*f;
	imap_t 	*imp;
{
#ifdef	DEBUG
	error("nimp %d\n", nimp);
#endif
	for (; imp; imp = imp->i_dxnext) {

		if (imp->i_flags & I_DELETE)
			continue;

		printonesym(f, imp);

		if (imp->i_flags & I_DIR && imp->i_dir) {
			fprintf(f, ">%c\n", 0);
			printsyms(f, imp->i_dir);
			fprintf(f, "<%c\n", 0);
		}
	}
}

#ifdef	__needed__
/*EXPORT BOOL*/
LOCAL BOOL
dirdiskonly(info, odep, odp)
	FINFO	*info;
	int	*odep;
	char	***odp;
{
	register char	**ep1;	   /* Directory entry pointer array (arch) */
	register char	**ep2 = 0; /* Directory entry pointer array (disk) */
	register char	*dp2;	   /* Directory names string from disk	   */
	register char	**oa = 0;  /* Only in arch pointer array	   */
	register char	**od = 0;  /* Only on disk pointer array	   */
	register int	i;
		int	ents1 = -1;
		int	ents2;
		int	dlen = 0;  /* # of entries only on disk		*/
		int	alen = 0;  /* # of entries only in arch		*/
		BOOL	diffs = FALSE;

	/*
	 * Old archives had only one nul at the end
	 * xheader.c already increments info->f_dirlen in this case
	 * but a newline may appear to be the last char.
	 * Note that we receicve the space from the xheader
	 * extract buffer.
	 */
	i = info->f_dirlen;
	if (info->f_dir[i-1] != '\0')
		info->f_dir[i-1] = '\0';	/* Kill '\n' */

	ep1 = sortdir(info->f_dir, &ents1);	/* from archive */
	dp2 = fetchdir(info->f_name, &ents2, 0, (ino_t **)0);
	if (dp2 == NULL) {
		diffs = TRUE;
		errmsg("Cannot read dir '%s'.\n", info->f_name);
		goto no_dircmp;
	}
	ep2 = sortdir(dp2, &ents2);		/* from disk */

	if (ents1 != ents2) {
		if (debug || verbose > 2) {
			error("Archive ents: %d Disk ents: %d '%s'\n",
					ents1, ents2, info->f_name);
		}
		diffs = TRUE;
	}

	if (cmpdir(ents1, ents2, ep1, ep2, NULL, NULL, &alen, &dlen) > 0)
		diffs = TRUE;

	oa = __malloc(alen * sizeof (char *), "dir diff array");
	od = __malloc(dlen * sizeof (char *), "dir diff array");
	cmpdir(ents1, ents2, ep1, ep2, oa, od, &alen, &dlen);

#ifdef	DEBUG
	for (i = 0; i < dlen; i++) {
		error("Only on disk '%s': '%s'\n",
				info->f_name, od[i] + 1);
	}
	for (i = 0; i < alen; i++) {
		error("Only in archive '%s': '%s'\n",
				info->f_name, oa[i] + 1);
	}
#endif

no_dircmp:
	if (odep)
		*odep = dlen;

	if (dp2)
		free(dp2);
	if (ep1)
		free(ep1);
	if (ep2)
		free(ep2);
	if (odp)
		*odp = od;
	else if (od)
		free(od);
	if (oa)
		free(oa);

	return (diffs);
}
#endif	/* __needed__ */
